fn main() {
    println!("Hello, world!");
}

mod rover {
    #[derive(Debug, PartialEq, Copy, Clone)]
    pub enum Facing { N, E, S, W }
    #[derive(Debug, PartialEq, Copy, Clone)]
    pub enum Rotation { Left, Right }

    #[derive(Debug, PartialEq, Copy, Clone)]
    pub struct Rover(pub i32, pub i32, pub Facing);

    fn velocity(direction : &Facing) -> (i32, i32) {
	match direction {
	    Facing::N  => ( 0, -1),
	    Facing::E  => ( 1,  0),
	    Facing::S  => ( 0,  1),
	    Facing::W  => (-1,  0)
	}
    }

    fn next_direction(direction : Facing) -> Facing {
	match direction {
	    Facing::N => Facing::W,
	    Facing::W => Facing::S,
	    Facing::S => Facing::E,
	    Facing::E => Facing::N,
	}
    }

    impl Rover {
	pub fn forward(&self) -> Rover {
	    let Rover(lon, lat, direction) = self;
	    let (xoff, yoff) = velocity(&direction);
	    Rover(lon + xoff as i32, lat + yoff as i32, *direction)
	}

	pub fn backward(&self) -> Rover {
	    let Rover(lon, lat, direction) = self;
	    let (xoff, yoff) = velocity(&direction);
	    Rover(lon - xoff as i32, lat - yoff as i32, *direction)
	}

	pub fn rotate(&self, rotation: Rotation) -> Rover {
	    let Rover(lon, lat, direction) = self;
	    match rotation {
		Rotation::Left => {
		    let new_d = next_direction(*direction);
		    Rover(*lon, *lat, new_d)
		},
		Rotation::Right => {
		    let new_d = next_direction(*direction);
		    let new_d = next_direction(new_d);
		    let new_d = next_direction(new_d);
		    Rover(*lon, *lat, new_d)
		}
	    }
	}
    }
}


mod driver {
    use super::rover;

    fn perform_command(command: &str, rover: &rover::Rover) -> rover::Rover {
	match command {
	    "f" => rover.forward(),
	    "b" => rover.backward(),
	    "l" => rover.rotate(rover::Rotation::Left),
	    "r" => rover.rotate(rover::Rotation::Right),
	    _ => rover.clone()
	}
    }

    pub fn execute(rover: &rover::Rover, commands: &str) -> rover::Rover {
	if commands.is_empty() {
	    rover.clone()
	} else {
	    let (now, later) = commands.split_at(1);
	    execute(&perform_command(now, rover), later)
	}
    }
}

#[cfg(test)]
mod tests {
    use super::{rover, driver};
    use super::rover::{Rover};

    #[test]
    fn move_north() {
	let r = Rover(1 as i32, 1 as i32, rover::Facing::N);
	assert_eq!(r.forward(),
		    Rover(1 as i32, 0 as i32, rover::Facing::N));

	let r =  Rover(1 as i32, 3 as i32, rover::Facing::N);
	assert_eq!(r.forward(),
		   Rover(1 as i32, 2 as i32, rover::Facing::N));
    }

    #[test]
    fn move_south() {
	let r = Rover(1 as i32, 1 as i32, rover::Facing::S);
	assert_eq!(r.forward(),
		   Rover(1 as i32, 2 as i32, rover::Facing::S));
    }

    #[test]
    fn move_west() {
	let r = Rover(1 as i32, 1 as i32, rover::Facing::W);
	assert_eq!(r.forward(),
		   Rover(0 as i32, 1 as i32, rover::Facing::W));
    }

    #[test]
    fn move_east() {
	let r = Rover(1 as i32, 1 as i32, rover::Facing::E);
	assert_eq!(r.forward(),
		   Rover(2 as i32, 1 as i32, rover::Facing::E));
    }

    #[test]
    fn move_backward() {
	let r = Rover(1 as i32, 1 as i32, rover::Facing::E);
	let r2 = r.forward();
	assert_eq!(r2.backward(), r);
    }

    #[test]
    fn spin_left() {
	let r = Rover(1 as i32, 1 as i32, rover::Facing::E);
	match r.rotate(rover::Rotation::Left) {
	    Rover(_, _, facing) => assert_eq!(facing, rover::Facing::N)
	}
    }
    #[test]
    fn spin_right() {
	let r = Rover(1 as i32, 1 as i32, rover::Facing::S);
	match r.rotate(rover::Rotation::Right) {
	    Rover(_, _, facing) => assert_eq!(facing, rover::Facing::W)
	}
    }

    #[test]
    fn drive() {
	let r = Rover(1 as i32, 1 as i32, rover::Facing::S);
	assert_eq!(driver::execute(&r, "ffrf"),
		   Rover(0 as i32, 3 as i32, rover::Facing::W));
    }
}
